----------------------------------------------------------------------------------------------------
-- SuperCup
----------------------------------------------------------------------------------------------------
-- Purpose: main class to handle the super cup.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class SuperCup
SuperCup = {}

SuperCup.STATES = {
    IDLE = 1,
    SETUP = 2,
    START = 3,
    RUN = 4,
    FINISHED = 5,
}

local SuperCup_mt = Class(SuperCup)

---Creates a new SuperCub instance.
---@return SuperCup
function SuperCup:new(mission, modDirectory, modName, farmId, baseNode)
    local self = setmetatable({}, SuperCup_mt)

    self.version = 1.0
    self.debug = false

    self.isServer = mission:getIsServer()
    self.isClient = mission:getIsClient()
    self.modDirectory = modDirectory
    self.modName = modName
    self.farmId = farmId
    self.baseNode = baseNode

    self.started = false

    self.currentPlayer = nil
    self.currentPlayerUserId = nil

    self.scores = {}
    self.checkpoints = {}
    self.vehicleList = {}
    self.vehiclesToLoad = {}
    self.superCupEnterReferenceNode = baseNode
    self.superCupExitReferenceNode = baseNode

    self.mission = mission

    self.coneHandler = ConeHandler:new(baseNode, modDirectory)
    self.trafficLight = TrafficLight:new(baseNode, self.isServer, self.isClient)

    self.hud = SuperCupHUD:new(modDirectory, mission.hud.gameInfoDisplay)
    self.activatable = SuperCupActivatable:new(self)

    self.backup_maxAttachDistanceSQ = AttacherJoints.MAX_ATTACH_DISTANCE_SQ
    self.backup_maxAttachAngle = AttacherJoints.MAX_ATTACH_ANGLE

    return self
end

---Called on delete.
function SuperCup:delete()
    removeTrigger(self.triggerNode)

    self.hud:delete()
    self.stateMachine:delete()
    self.coneHandler:delete()
    self.trafficLight:delete()

    for _, checkpoint in ipairs(self.checkpoints) do
        checkpoint:delete()
    end
    self.checkpoints = {}

    for _, vehicle in ipairs(self.vehicleList) do
        if vehicle ~= nil and not vehicle.isDeleted then
            g_currentMission:removeVehicle(vehicle, false)
            vehicle:delete()
        end
    end

    self.vehicleList = {}
    self.vehiclesToLoad = {}
end

---Load the super cup from the XML.
function SuperCup:loadFromXML(xmlFile, rootXMLTag)
    self.superCupEnterReferenceNode = I3DUtil.indexToObject(self.baseNode, getXMLString(xmlFile, rootXMLTag .. ".enterReferenceNode"))
    self.superCupExitReferenceNode = I3DUtil.indexToObject(self.baseNode, getXMLString(xmlFile, rootXMLTag .. ".exitReferenceNode"))
    self.triggerNode = I3DUtil.indexToObject(self.baseNode, getXMLString(xmlFile, rootXMLTag .. "#triggerNode"))
    assert(self.triggerNode ~= nil, "Supercup needs a start trigger!")

    addTrigger(self.triggerNode, "playerEnterTriggerCallback", self)

    self.hud:loadFromXML(xmlFile, rootXMLTag)
    self.coneHandler:loadFromXML(xmlFile, rootXMLTag)
    self.trafficLight:loadFromXML(xmlFile, rootXMLTag)
    self:loadSuperCupCheckpointsFromXML(xmlFile, rootXMLTag)
    self:loadSuperCupVehiclesFromXML(xmlFile, rootXMLTag)

    local function updateVehicleList(state)
        state.vehicleList = self.vehicleList
    end

    local function setAttachRangeLimit()
        AttacherJoints.MAX_ATTACH_DISTANCE_SQ = 0.02
        AttacherJoints.MAX_ATTACH_ANGLE = 0.085505
    end

    local function resetAttachRangeLimit()
        AttacherJoints.MAX_ATTACH_DISTANCE_SQ = self.backup_maxAttachDistanceSQ
        AttacherJoints.MAX_ATTACH_ANGLE = self.backup_maxAttachAngle
    end

    self.stateMachine = StateEngine:new()
                                   :add(SuperCup.STATES.IDLE, AbstractState)
                                   :add(SuperCup.STATES.SETUP, SetupState, { self.mission, self.coneHandler, self.checkpoints, self.vehiclesToLoad, self.vehicleList })
                                   :add(SuperCup.STATES.START, StartState, { self.trafficLight, self.hud })
                                   :add(SuperCup.STATES.RUN, RunState, { self.coneHandler, self.checkpoints, self.hud })
                                   :add(SuperCup.STATES.FINISHED, FinishState, { self, self.vehicleList, self.trafficLight, self.scores, self.hud })
                                   :addEntryAction(SuperCup.STATES.START, setAttachRangeLimit)
                                   :addEntryAction(SuperCup.STATES.FINISHED, resetAttachRangeLimit)
                                   :addEntryAction(SuperCup.STATES.FINISHED, updateVehicleList)
                                   :build()
end

---Called when mission is being saved with our own xml file.
function SuperCup:saveToXMLFile()
    local xmlFile = self:getXMLFile(self:getFilePath())
    if xmlFile ~= nil then
        local dateNow = getDate("%Y-%m-%d")

        local d = 0
        while true do
            local key = ("superCup.scores(%d)"):format(d)
            if not hasXMLProperty(xmlFile, key) then
                break
            end

            if getXMLString(xmlFile, key .. "#date") == dateNow then
                break
            end
            d = d + 1
        end

        local baseKey = ("superCup.scores(%d)"):format(d)
        setXMLString(xmlFile, baseKey .. "#date", dateNow)
        for i, score in ipairs(self.scores) do
            local key = (("%s.score(%d)"):format(baseKey, i - 1))
            setXMLString(xmlFile, key .. "#time", score.time)
            setXMLInt(xmlFile, key .. "#penalties", score.penalties)
            setXMLString(xmlFile, key .. "#player", score.playerNickname)
        end

        saveXMLFile(xmlFile)
        delete(xmlFile)
    end
end

function SuperCup:getXMLFile(path)
    local xmlFile
    if fileExists(path) then
        xmlFile = loadXMLFile("superCupXML", path)
    else
        xmlFile = createXMLFile("superCupXML", path, "superCup")
    end

    return xmlFile
end

function SuperCup:getFilePath()
    return Utils.getFilename("superCup.xml", getUserProfileAppPath())
end

---Called on update.
function SuperCup:update(dt)
    self.stateMachine:update(dt)

    if self.started then
        self.mission:addActivatableObject(self.activatable)
    end
end

function SuperCup:draw()
    self.hud:draw()
end

function SuperCup:onYesNoSuperCupDialog(isYes)
    if isYes then
        self:startSuperCup()
    end
end

function SuperCup:setIsSuperCupStarted(isStarted)
    if isStarted then
        self:startSuperCup()
    else
        self:stopSuperCup()
    end
end

function SuperCup:startSuperCup()
    self.currentPlayer = self.mission.player
    self.currentPlayerUserId = self.mission.playerUserId
    self.started = true

    if not self.stateMachine:isInState(SuperCup.STATES.SETUP) then
        self.stateMachine:transition(SuperCup.STATES.SETUP)
    end
end

function SuperCup:stopSuperCup()
    local player = self.currentPlayer
    self.currentPlayer = nil
    self.currentPlayerUserId = nil
    self.started = false
    self.mission:removeActivatableObject(self.activatable)

    if not self.stateMachine:isInState(SuperCup.STATES.FINISHED) then
        self.stateMachine:transition(SuperCup.STATES.FINISHED)
    end

    local x, y, z = getWorldTranslation(self.superCupExitReferenceNode)
    player:moveToAbsolute(x, y, z)
end

function SuperCup:loadSuperCupCheckpointsFromXML(xmlFile, rootXMLTag)
    local i = 0
    while true do
        local key = ("%s.checkpoints.checkpoint(%d)"):format(rootXMLTag, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local className = Utils.getNoNil(getXMLString(xmlFile, key .. "#className"), "Checkpoint")

        local checkpointClass = SuperCupUtil.getClassObject(className)
        if checkpointClass ~= nil then
            local checkpoint = checkpointClass:new(self.baseNode, self.isServer, self.isClient)

            if checkpoint:loadFromXML(xmlFile, key) then
                table.insert(self.checkpoints, checkpoint)
            else
                checkpoint:delete()
            end
        end

        i = i + 1
    end
end

function SuperCup:loadSuperCupVehiclesFromXML(xmlFile, rootXMLTag)
    if not self.isServer then
        return
    end

    local i = 0
    while true do
        local key = ("%s.vehicles.vehicle(%d)"):format(rootXMLTag, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local filename = getXMLString(xmlFile, key .. "#filename")
        local spawnNode = I3DUtil.indexToObject(self.baseNode, getXMLString(xmlFile, key .. "#spawnNode"))

        if filename == nil or spawnNode == nil then
            g_logManager:xmlWarning("Could not find vehicle file or spawn node!")
            break
        end

        local rotation = math.rad(Utils.getNoNil(getXMLFloat(xmlFile, key .. "#rotation"), 0))

        table.insert(self.vehiclesToLoad, { node = spawnNode, filename = Utils.getFilename(filename, self.modDirectory), rotation = rotation, farmId = self.farmId , baseNode = self.baseNode})

        i = i + 1
    end
end

function SuperCup:playerEnterTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    local player = self.mission.player
    if onEnter and player ~= nil and otherId == player.rootNode then
        if not self.started then
            g_gui:showYesNoDialog({
                dialogType = DialogElement.TYPE_QUESTION,
                text = g_i18n:getText("info_participate"),
                callback = self.onYesNoSuperCupDialog,
                target = self
            })
        end
    end
end
