----------------------------------------------------------------------------------------------------
-- Timer
----------------------------------------------------------------------------------------------------
-- Purpose: Timer class to handle actions that require timers to avoid blobs of code doing the same thing.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class Timer
Timer = {}

local Timer_mt = Class(Timer)

---Creates a new Timer instance.
---@return Timer
function Timer:new(delay, actionListener, hostObject)
    local self = setmetatable({}, Timer_mt)

    self.delay = delay

    -- Time to wait after the timer is started before calling listeners.
    self.initialDelay = delay

    self.elapsedDelay = delay
    self.running = false
    self.repeats = false
    self.isStopWatch = false

    self.actionListeners = {}
    SuperCupUtil.addToListIfNotExists(actionListener, self.actionListeners)

    self.hostObject = hostObject or g_currentMission

    assert(self.hostObject.addUpdateable ~= nil, "Host object does not have the addUpdateable function!")
    assert(self.hostObject.removeUpdateable ~= nil, "Host object does not have the removeUpdateable function!")

    return self
end

function Timer.schedule(delay, actionListener)
    local t = Timer:new(delay, actionListener)

    t:start()

    return t
end


function Timer:delete()
    self:clear()
end

---Clears the timer.
function Timer:clear()
    self.running = false
    self.elapsedDelay = 0
    self.hostObject:removeUpdateable(self)
end

---Returns the set delay.
function Timer:getDelay()
    return self.delay
end

---Sets the delay.
function Timer:setDelay(delay)
    self.delay = delay
end

---Returns true when the repeating flag is set, false otherwise.
function Timer:isRepeats()
    return self.repeats
end

---Set repeating flag.
function Timer:setIsRepeats(isRepeats)
    self.repeats = isRepeats
end

---Set stopwatch flag.
function Timer:setIsStopWatch(isStopWatch)
    self.isStopWatch = isStopWatch
    self:setDelay(0)
end

---Starts the timer. When the delay is reached it will fire and event to the listeners.
function Timer:start()
    self.hostObject:addUpdateable(self)
    self.running = true
    self.elapsedDelay = self.delay
end

---Stops the timer. The listeners will not get updates.
function Timer:stop()
    self.hostObject:removeUpdateable(self)
    self.running = false
end

---Stops and starts the timer again.
function Timer:restart()
    self:stop()
    self:start()
end

---Clear the timer and call the listeners.
function Timer:fire()
    self:clear()

    for _, actionListener in ipairs(self.actionListeners) do
        actionListener()
    end
end

---Returns true when running, false otherwise.
function Timer:isRunning()
    return self.running
end

---Gets the elapsed time.
function Timer:getElapsedTime()
    return self.elapsedDelay
end

---Called on update frame by host object.
function Timer:update(dt)
    if self.running then
        if self.isStopWatch then
            self.elapsedDelay = self.elapsedDelay + dt
        else
            self.elapsedDelay = self.elapsedDelay - dt

            if self.elapsedDelay <= 0 then
                self:fire()

                if self.repeats then
                    self:restart()
                end
            end
        end
    end
end

---Adds a new action listener to the list.
function Timer:addActionListener(actionListener)
    SuperCupUtil.addToListIfNotExists(actionListener, self.actionListeners)
end

---Removes the given action listener from the list.
function Timer:removeActionListener(actionListener)
    ListUtil.removeElementFromList(self.actionListeners, actionListener)
end

---Returns the action listeners.
function Timer:getActionListeners()
    return self.actionListeners
end
