----------------------------------------------------------------------------------------------------
-- ActionRequiredCheckpoint
----------------------------------------------------------------------------------------------------
-- Purpose: checkpoint that requires an action to be fulfilled in order to pass.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class ActionRequiredCheckpoint
ActionRequiredCheckpoint = {}

local ActionRequiredCheckpoint_mt = Class(ActionRequiredCheckpoint, Checkpoint)

---Creates a new ActionRequiredCheckpoint instance.
---@return ActionRequiredCheckpoint
function ActionRequiredCheckpoint:new(nodeId, isServer, isClient, action)
    local self = Checkpoint:new(nodeId, isServer, isClient, ActionRequiredCheckpoint_mt)

    self.action = action
    self.object = nil

    return self
end

function ActionRequiredCheckpoint:delete()
    ActionRequiredCheckpoint:superClass().delete(self)

    if self.action ~= nil then
        self.action:delete()
    end
end

function ActionRequiredCheckpoint:reset()
    ActionRequiredCheckpoint:superClass().reset(self)

    if self.action ~= nil then
        self.action:reset()
    end
end

function ActionRequiredCheckpoint:loadFromXML(xmlFile, baseKey)
    if not ActionRequiredCheckpoint:superClass().loadFromXML(self, xmlFile, baseKey) then
        return false
    end

    local actionClassName = Utils.getNoNil(getXMLString(xmlFile, baseKey .. "#actionClass"), "CheckpointAction")

    local actionClass = SuperCupUtil.getClassObject(actionClassName)
    if actionClass ~= nil then
        local action = actionClass:new(self)

        if action:loadFromXML(xmlFile, baseKey) then
            self:setRequiredAction(action)
        end
    end

    return true
end

---Sets the action that needs to executed in order to pass the checkpoint.
---@param action function
function ActionRequiredCheckpoint:setRequiredAction(action)
    self.action = action
end

function ActionRequiredCheckpoint:hasPassedCheckpoint()
    if self.action ~= nil then
        self.action:validate({ self.object })

        if not self.action:isFulfilled() then
            return false
        end
    end

    if not ActionRequiredCheckpoint:superClass().hasPassedCheckpoint(self) then
        return false
    end

    return true
end

function ActionRequiredCheckpoint:checkpointTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    ActionRequiredCheckpoint:superClass().checkpointTriggerCallback(self, triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)

    -- Allow action to prepare before active state.
    if self.action ~= nil and self.action.checkpointTriggerCallback ~= nil then
        self.action:checkpointTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    end

    if self.isActive then
        if otherId ~= 0 then
            if onEnter then
                local object = g_currentMission:getNodeObject(otherId)
                self.object = object
            elseif onLeave then
                self.object = nil
            end
        end
    end
end
