----------------------------------------------------------------------------------------------------
-- ConeHandler
----------------------------------------------------------------------------------------------------
-- Purpose: Class to handle all cones in an area.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class ConeHandler
ConeHandler = {}

local ConeHandler_mt = Class(ConeHandler)

---Creates a new ConeHandler instance.
---@return ConeHandler
function ConeHandler:new(nodeId, modDirectory, mt)
    local self = setmetatable({}, mt or ConeHandler_mt)

    self.nodeId = nodeId
    self.modDirectory = modDirectory
    self.cones = {}
    self.activeCones = {}

    self.actionListeners = {}

    return self
end

function ConeHandler:delete()
    for _, cone in ipairs(self.cones) do
        self:destroyCone(cone)
    end
end

function ConeHandler:loadFromXML(xmlFile, baseKey)
    local i = 0
    while true do
        local key = ("%s.cones.cone(%d)"):format(baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#node"))

        if node ~= nil then
            self:spawnConeAt(node)
        end

        i = i + 1
    end
end

function ConeHandler:add(cone)
    table.insert(self.cones, cone)
end

function ConeHandler:spawnCone(x, y, z)
    local cone = Cone:new(g_server ~= nil, g_client ~= nil)

    cone:load(self.modDirectory .. "arena/cone.i3d")
    cone:link(self.nodeId)
    cone:setPosition(x, y, z)
    self:add(cone)

    return cone
end

function ConeHandler:spawnConeAt(node)
    local x, y, z = getWorldTranslation(node)
    self:spawnCone(x, y, z)
end

function ConeHandler:destroyCone(cone)
    cone:delete()
end

function ConeHandler:coneLostBall(cone)
    self.lostBallCount = self.lostBallCount + 1
    ListUtil.removeElementFromList(self.activeCones, cone)

    for _, actionListener in ipairs(self.actionListeners) do
        actionListener(cone, self.lostBallCount)
    end
end

function ConeHandler:getLostBallCount()
    return self.lostBallCount
end

function ConeHandler:reset()
    for _, cone in ipairs(self.cones) do
        cone:reset()
    end

    self.lostBallCount = 0
    self.activeCones = ListUtil.copyTable(self.cones)
end

function ConeHandler:update(dt)
    for _, cone in ipairs(self.activeCones) do
        cone:update(dt)

        if cone.lostBall then
            self:coneLostBall(cone)
        end
    end
end

---Adds a new action listener to the list.
function ConeHandler:addActionListener(actionListener)
    SuperCupUtil.addToListIfNotExists(actionListener, self.actionListeners)
end

---Removes the given action listener from the list.
function ConeHandler:removeActionListener(actionListener)
    ListUtil.removeElementFromList(self.actionListeners, actionListener)
end

---Returns the action listeners.
function ConeHandler:getActionListeners()
    return self.actionListeners
end
