----------------------------------------------------------------------------------------------------
-- Seesaw
----------------------------------------------------------------------------------------------------
-- Purpose: Class to control the seesaw.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class Seesaw
Seesaw = {}
Seesaw.ANGLE_THRESHOLD = 1 -- deg

local Seesaw_mt = Class(Seesaw)

InitObjectClass(Seesaw, "Seesaw")

function Seesaw:new(nodeId, isServer, isClient)
    local self = setmetatable({}, Seesaw_mt)

    self.nodeId = nodeId
    self.ramps = {}

    self.isServer = isServer
    self.isClient = isClient

    registerObjectClassName(self, "Seesaw")

    return self
end

function Seesaw:delete()
    if self.beaconLight ~= nil and self.beaconLight.filename ~= nil then
        g_i3DManager:releaseSharedI3DFile(self.beaconLight.filename, nil, true)
    end

    unregisterObjectClassName(self)
end

function Seesaw:loadFromXML(xmlFile, baseKey)
    self.rootNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".seesaw#rootNode"))
    assert(self.rootNode ~= nil, "Seesaw requires a rootNode!")
    assert(getRigidBodyType(self.rootNode) ~= "NoRigidBody", "Seesaw rootNode needs to be a rigid body!")

    local i = 0
    while true do
        local key = ("%s.seesaw.ramps.ramp(%d)"):format(baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local ramp = {}
        if self:loadRampFromXML(xmlFile, key, ramp) then
            table.insert(self.ramps, ramp)
        end

        i = i + 1
    end

    self:loadBeaconLightFromXML(xmlFile, baseKey)
    self:removeFromPhysics()
end

function Seesaw:loadRampFromXML(xmlFile, baseKey, ramp)
    local node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. "#node"))
    local jointTransform = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. "#jointTransform"))

    if node == nil or jointTransform == nil then
        return false
    end

    ramp.node = node
    ramp.orgTranslation = { getTranslation(node) }
    ramp.orgRotation = { getRotation(node) }
    ramp.maxAngle = Utils.getNoNilRad(getXMLFloat(xmlFile, baseKey .. "#maxAngle"), math.rad(10))
    ramp.jointTransform = jointTransform
    ramp.jointIndex = 0

    return true
end

function Seesaw:loadBeaconLightFromXML(xmlFile, baseKey)
    local node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".seesaw.beaconLight#node"))
    if node ~= nil then
        local lightXmlFilename = getXMLString(xmlFile, baseKey .. ".seesaw.beaconLight#filename")

        if lightXmlFilename ~= nil then
            lightXmlFilename = Utils.getFilename(lightXmlFilename)
            local lightXmlFile = loadXMLFile("beaconLightXML", lightXmlFilename)

            if lightXmlFile ~= nil and lightXmlFile ~= 0 then
                local i3dFilename = getXMLString(lightXmlFile, "beaconLight.filename")

                if i3dFilename ~= nil then
                    local i3dNode = g_i3DManager:loadSharedI3DFile(i3dFilename, nil, false, false, false)

                    if i3dNode ~= nil and i3dNode ~= 0 then
                        local rootNode = I3DUtil.indexToObject(i3dNode, getXMLString(lightXmlFile, "beaconLight.rootNode#node"))
                        local rotatorNode = I3DUtil.indexToObject(i3dNode, getXMLString(lightXmlFile, "beaconLight.rotator#node"))
                        local speed = getXMLFloat(lightXmlFile, "beaconLight.rotator#speed") or 0.015
                        local lightNode = I3DUtil.indexToObject(i3dNode, getXMLString(lightXmlFile, "beaconLight.light#node"))
                        local lightShaderNode = I3DUtil.indexToObject(i3dNode, getXMLString(lightXmlFile, "beaconLight.light#shaderNode"))
                        local intensity = getXMLFloat(lightXmlFile, "beaconLight.light#intensity") or 1000
                        local realLightNode = I3DUtil.indexToObject(i3dNode, getXMLString(lightXmlFile, "beaconLight.realLight#node"))

                        if rootNode ~= nil and (lightNode ~= nil or lightShaderNode ~= nil) then
                            link(node, rootNode)
                            setTranslation(rootNode, 0, 0, 0)

                            local light = {
                                filename = i3dFilename,
                                rootNode = rootNode,
                                rotatorNode = rotatorNode,
                                lightNode = lightNode,
                                lightShaderNode = lightShaderNode,
                                realLightNode = realLightNode,
                                speed = speed,
                                intensity = intensity
                            }

                            if realLightNode ~= nil then
                                light.defaultColor = {
                                    getLightColor(realLightNode)
                                }

                                setVisibility(realLightNode, false)
                            end

                            if lightNode ~= nil then
                                setVisibility(lightNode, false)
                            end

                            if lightShaderNode ~= nil then
                                local _, y, z, w = getShaderParameter(lightShaderNode, "lightControl")
                                setShaderParameter(lightShaderNode, "lightControl", 0, y, z, w, false)
                            end

                            if light.speed > 0 then
                                local rot = math.random(0, math.pi * 2)

                                if light.rotatorNode ~= nil then
                                    setRotation(light.rotatorNode, 0, rot, 0)
                                end
                            end

                            self.beaconLight = light
                        end

                        delete(i3dNode)
                    end
                end

                delete(lightXmlFile)
            end
        end
    end

    self.beaconLightsActive = false
end

function Seesaw:createRampJoint(ramp)
    local constructor = JointConstructor:new()

    constructor:setActors(self.rootNode, ramp.node)
    constructor:setJointTransforms(ramp.jointTransform, ramp.jointTransform)

    local springForce = 10500
    local springDamping = 5500
    constructor:setEnableCollision(false)
    constructor:setRotationLimitSpring(springForce, springDamping, springForce, springDamping, springForce, springDamping)
    constructor:setTranslationLimitSpring(springForce, springDamping, springForce, springDamping, springForce, springDamping)

    constructor:setTranslationLimit(0, true, 0, 0)
    constructor:setTranslationLimit(1, true, 0, 0)
    constructor:setTranslationLimit(2, true, 0, 0)

    constructor:setRotationLimit(0, 0, 0)
    constructor:setRotationLimit(1, 0, 0)
    constructor:setRotationLimit(2, -ramp.maxAngle, ramp.maxAngle)

    return constructor:finalize()
end

function Seesaw:reset()
    self:removeFromPhysics()

    for _, ramp in ipairs(self.ramps) do
        setTranslation(ramp.node, unpack(ramp.orgTranslation))
        setRotation(ramp.node, unpack(ramp.orgRotation))
    end

    self:addToPhysics()

    for _, ramp in ipairs(self.ramps) do
        setRotation(ramp.node, 0, 0, ramp.maxAngle)
        setJointFrame(ramp.jointIndex, 0, ramp.jointTransform)
    end
end

function Seesaw:addToPhysics()
    if not self.isAddedToPhysics then
        addToPhysics(self.rootNode)

        for _, ramp in ipairs(self.ramps) do
            addToPhysics(ramp.node)
            if self.isServer then
                ramp.jointIndex = self:createRampJoint(ramp)
            end
        end

        self.isAddedToPhysics = true
    end
end

function Seesaw:removeFromPhysics()
    removeFromPhysics(self.rootNode)

    for _, ramp in ipairs(self.ramps) do
        removeFromPhysics(ramp.node)
        if self.isServer then
            ramp.jointIndex = 0
        end
    end

    self.isAddedToPhysics = false
end

function Seesaw:isHorizontal()
    local ramp1, ramp2 = unpack(self.ramps)

    local _, _, rz1 = getRotation(ramp1.node)
    local _, _, rz2 = getRotation(ramp2.node)

    if math.abs(math.deg(rz1)) <= Seesaw.ANGLE_THRESHOLD
        and math.abs(math.deg(rz2)) <= Seesaw.ANGLE_THRESHOLD then
        return true
    end

    return false
end

function Seesaw:updateBeaconLight(dt)
    if self.beaconLightsActive then
        local beaconLight = self.beaconLight
        if beaconLight.rotatorNode ~= nil then
            rotate(beaconLight.rotatorNode, 0, beaconLight.speed * dt, 0)
        end
    end
end

function Seesaw:setBeaconLightVisibility(visibility, force, noEventSend)
    if visibility ~= self.beaconLightsActive or force then
        if noEventSend == nil or noEventSend == false then
            -- todo: if MP support.
        end

        self.beaconLightsActive = visibility
        local beaconLight = self.beaconLight
        if beaconLight ~= nil then
            local realBeaconLights = g_gameSettings:getValue("realBeaconLights")

            if realBeaconLights and beaconLight.realLightNode ~= nil then
                setVisibility(beaconLight.realLightNode, visibility)
            end

            if beaconLight.lightNode ~= nil then
                setVisibility(beaconLight.lightNode, visibility)
            end

            if beaconLight.lightShaderNode ~= nil then
                local value = visibility and 1 * beaconLight.intensity or 0
                local _, y, z, w = getShaderParameter(beaconLight.lightShaderNode, "lightControl")
                setShaderParameter(beaconLight.lightShaderNode, "lightControl", value, y, z, w, false)
            end
        end
    end

    return true
end
