----------------------------------------------------------------------------------------------------
-- FSM
----------------------------------------------------------------------------------------------------
-- Purpose: Finite-state machine implementation.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class FSM
---@field public state AbstractState
---@field public states AbstractState[]
FSM = {}

---@type number The any state.
FSM.ANY_STATE = -1

local FSM_mt = Class(FSM)

---Creates a new finite state machine instance
---@param initialState AbstractState
---@param states AbstractState[]
---@param custom_mt table
---@return FSM
function FSM:new(initialState, states, custom_mt)
    local self = setmetatable({}, custom_mt or FSM_mt)

    self.initialState = initialState
    self.state = initialState

    -- This includes all the states
    -- Each state has an identifier which can be called by getId()
    self.states = states or {}

    return self
end

function FSM:delete()
    if self.state ~= nil then
        self.state:delete()
    end

    self.state = nil
    self.states = {}
end

---Transitions to new state
---@param toStateId number the state id we transition to.
function FSM:transition(toStateId, params)
    assert(self.states[toStateId] ~= nil)
    local nextState = self.states[toStateId]
    local fromState = self.state

    -- We exit the previous state
    fromState:onExit()

    self:setCurrentState(nextState)

    -- We enter the new state
    nextState:onEnter(unpack(params or {}))
end

---Current state needs update.
function FSM:requestStateUpdate()
    local state = self:getCurrentState()

    -- We call the state update actions.
    if state.updateActions ~= nil then
        for _, action in ipairs(state.updateActions) do
            action(state)
        end
    end
end

---Sets the current state
---@param state AbstractState
function FSM:setCurrentState(state)
    self.state = state
end

---Returns the current active state
function FSM:getCurrentState()
    return self.state
end

---Returns true when the state equals, false otherwise
function FSM:isInState(id)
    return id == self.state.id
end

---Sets the initial state
---@param state AbstractState
function FSM:setInitialState(state)
    self.initialState = state
end

---Returns the initial state of the state machine
---@return AbstractState
function FSM:getInitialState()
    return self.initialState
end

---Resets the state machine to the initial state
function FSM:reset()
    local state = self:getInitialState()
    self:transition(state)
end

---Sets the possible states
---@param states AbstractState[]
function FSM:setStates(states)
    self.states = states
end

---Updates the current state
---@param dt number
function FSM:update(dt)
    self.state:onUpdate(dt)
end

---Draws the current state
function FSM:draw()
    self.state:onDraw()
end
