----------------------------------------------------------------------------------------------------
-- StateEngine
----------------------------------------------------------------------------------------------------
-- Purpose: Engine to build the state machine with.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class StateEngine
StateEngine = {}

local StateEngine_mt = Class(StateEngine)

---Creates a new instance of the StateEngine.
---@return StateEngine
function StateEngine:new()
    local self = setmetatable({}, StateEngine_mt)

    self.stateMachine = FSM:new()
    self.states = {}

    return self
end

---Adds a state to the engine.
---@param stateId number
---@param state AbstractState
function StateEngine:add(stateId, state, params)
    assert(self.states[stateId] == nil, "State " .. stateId .. " already exists!")
    self.states[stateId] = state:new(stateId, self.stateMachine, unpack(params or {}))
    return self
end

---Adds an entry action to the given state.
---@param stateId number
---@param action function
function StateEngine:addEntryAction(stateId, action)
    assert(self.states[stateId] ~= nil, "State must exist!")
    self.states[stateId]:addEntryAction(action)
    return self
end

---Adds an exit action to the given state.
---@param stateId number
---@param action function
function StateEngine:addExitAction(stateId, action)
    assert(self.states[stateId] ~= nil, "State must exist!")
    self.states[stateId]:addExitAction(action)
    return self
end

---Adds an update action to the given state.
---@param stateId number
---@param action function
function StateEngine:addUpdateAction(stateId, action)
    assert(self.states[stateId] ~= nil, "State must exist!")
    self.states[stateId]:addUpdateAction(action)
    return self
end

--- Create an instance of a FSM that uses this state engine.
--- The initial state of the FSM will be the initialState specified here.
---@return FSM
function StateEngine:build()
    if #self.states == 0 then
        log("Please add states first!")
        return nil
    end

    self.stateMachine:setStates(self.states)
    local initialState = self.states[1]
    self.stateMachine:setInitialState(initialState)
    self.stateMachine:setCurrentState(initialState)

    return self.stateMachine
end
