----------------------------------------------------------------------------------------------------
-- AbstractState
----------------------------------------------------------------------------------------------------
-- Purpose: Abstract state for the state machine.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class AbstractState
---@field public id number
AbstractState = {}

local AbstractState_mt = Class(AbstractState)

---Creates a new abstract state.
---@param id number
---@param stateMachine FSM
---@param custom_mt table
---@return AbstractState
function AbstractState:new(id, stateMachine, custom_mt)
    local self = setmetatable({}, custom_mt or AbstractState_mt)

    self.id = id
    self.stateMachine = stateMachine

    self.entryActions = {}
    self.exitActions = {}
    self.updateActions = {}

    return self
end

---Called on delete.
function AbstractState:delete()
    self.stateMachine = nil
    self.entryActions = {}
    self.exitActions = {}
    self.updateActions = {}
end

---Gets the current state id.
---@return number
function AbstractState:getId()
    return self.id
end

---Called on entry state.
function AbstractState:onEnter(...)
    for _, entryAction in ipairs(self.entryActions) do
        entryAction(self, ...)
    end
end

---Called on exit state.
function AbstractState:onExit(...)
    for _, exitAction in ipairs(self.exitActions) do
        exitAction(self, ...)
    end
end

---Called on each update frame.
---@param dt number delta time.
---@return number
function AbstractState:onUpdate(dt)
end

---Called on each draw frame.
---@return number
function AbstractState:onDraw()
end

---Adds onEntry actions.
---@param action function
function AbstractState:addEntryAction(action)
    SuperCupUtil.addToListIfNotExists(action, self.entryActions)
end

---Adds onExit actions.
---@param action function
function AbstractState:addExitAction(action)
    SuperCupUtil.addToListIfNotExists(action, self.exitActions)
end

---Adds state update actions.
---@param action function
function AbstractState:addUpdateAction(action)
    SuperCupUtil.addToListIfNotExists(action, self.updateActions)
end
