----------------------------------------------------------------------------------------------------
-- RunState
----------------------------------------------------------------------------------------------------
-- Purpose: Run state for the state machine.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class RunState
---@field public id number
RunState = {}
RunState.PENALTY_TIME = 5000 --ms

local RunState_mt = Class(RunState, AbstractState)

---Creates a new run state.
---@param id number
---@param stateMachine FSM
---@param custom_mt table
---@return RunState
function RunState:new(id, stateMachine, coneHandler, checkpoints, hud, custom_mt)
    local self = AbstractState:new(id, stateMachine, custom_mt or RunState_mt)

    self.action = function(lastCone, count)
        self:displayLostBallCount(lastCone, count)
    end

    self.coneHandler = coneHandler
    self.checkpoints = checkpoints
    self.hud = hud

    self.queue = Queue:new()
    self.timer = Timer:new()
    self.timer:setIsStopWatch(true)

    return self
end

---Called on delete.
function RunState:delete()
    RunState:superClass().delete(self)
end

---Called on enter.
function RunState:onEnter(runVehicle)
    RunState:superClass().onEnter(self)
    runVehicle:setAllowDriveSuperCup(true)

    self.coneHandler:addActionListener(self.action)

    self.hud:setIsActive(true)
    self.hud:setTimer(self.timer)
    self.hud:setPenaltyPoints(0)

    self.timer:start()

    self.queue = Queue.fromList(ListUtil.copyTable(self.checkpoints))
    self.currentCheckpoint = self.queue:pop()
    self.currentCheckpoint:setIsActive(true)
end

---Called on exit.
function RunState:onExit()
    RunState:superClass().onExit(self)
    self.coneHandler:removeActionListener(self.action)
    self.timer:stop()
end

function RunState:onUpdate(dt)
    self.coneHandler:update(dt)

    if self.currentCheckpoint ~= nil then
        if self.currentCheckpoint:hasPassedCheckpoint() then
            --log("Awesome, you passed a checkpoint!")
            self.currentCheckpoint:setIsActive(false)
            self.currentCheckpoint = self.queue:pop()
            if self.currentCheckpoint ~= nil then
                self.currentCheckpoint:setIsActive(true)
            end
        end
    else

        --Todo: set active player on the super cup.
        local nickname = g_currentMission.userManager:getUserByUserId(g_currentMission.playerUserId):getNickname()
        local score = { playerNickname = nickname, time = SuperCupUtil.millisecondsToClockString(self.timer:getElapsedTime()), penalties = self.coneHandler:getLostBallCount() }
        self.stateMachine:transition(SuperCup.STATES.FINISHED, { score })
    end
end

function RunState:displayLostBallCount(lastCone, count)
    self.timer.elapsedDelay = self.timer.elapsedDelay + RunState.PENALTY_TIME
    self.hud:setPenaltyPoints(count)
end
