----------------------------------------------------------------------------------------------------
-- SetupState
----------------------------------------------------------------------------------------------------
-- Purpose: Setup state for the state machine.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class SetupState
---@field public id number
SetupState = {}

local SetupState_mt = Class(SetupState, AbstractState)

---Creates a new setup state.
---@param id number
---@param stateMachine FSM
---@param custom_mt table
---@return SetupState
function SetupState:new(id, stateMachine, mission, coneHandler, checkpoints, vehicles, vehicleList, custom_mt)
    local self = AbstractState:new(id, stateMachine, custom_mt or SetupState_mt)

    self.isServer = mission:getIsServer()
    self.mission = mission
    self.coneHandler = coneHandler
    self.checkpoints = checkpoints
    self.vehicles = vehicles
    self.vehicleList = vehicleList
    self.runVehicle = nil

    self.spawnTimer = nil

    return self
end

---Called on delete.
function SetupState:delete()
    SetupState:superClass().delete(self)

    if self.spawnTimer ~= nil then
        self.spawnTimer:delete()
    end
end

---Called on enter.
function SetupState:onEnter()
    SetupState:superClass().onEnter(self)


    --Reset checkpoints before vehicle spawn because of some checkpoints being dependent on vehicles.
    for _, checkpoint in ipairs(self.checkpoints) do
        checkpoint:reset()
    end

    self:spawn()
end

---Called on exit.
function SetupState:onExit()
    SetupState:superClass().onExit(self)

    --Clear spawn timer in case we do an abrupt state change.
    if self.spawnTimer ~= nil then
        self.spawnTimer:clear()
    end
end

function SetupState:spawn()
    self:spawnVehicles(function()
        self.coneHandler:reset()

        self.mission:enterVehicleWithPlayer(self.runVehicle, self.mission.player)

        --Delay so that the vehicles can spawn.
        self.spawnTimer = Timer.schedule(2000, function()
            self.stateMachine:transition(SuperCup.STATES.START, { self.runVehicle })
        end)
    end)
end

function SetupState:spawnVehicles(callBack)
    local vehiclesToSpawn = Queue.fromList(ListUtil.copyTable(self.vehicles))

    if vehiclesToSpawn:isEmpty() then
        callBack()
        return
    end

    local function loadFinished(self, vehicle, state, args)
        if vehicle == nil then
            g_logManager:xmlWarning("Could not spawn super cup vehicle!")
            return
        end

        if vehicle.getIsEnterable ~= nil and vehicle:getIsEnterable() then
            self.runVehicle = vehicle
        end

        vehicle.isVehicleSaved = false
        table.insert(self.vehicleList, vehicle)

        local queue = unpack(args)
        if not queue:isEmpty() then
            self:spawnVehicleAtNode(vehiclesToSpawn:pop(), loadFinished, vehiclesToSpawn)
        else
            callBack()
        end
    end

    self:spawnVehicleAtNode(vehiclesToSpawn:pop(), loadFinished, vehiclesToSpawn)
end

function SetupState:spawnVehicleAtNode(vehicle, loadFinished, queue)
    local x, y, z = getWorldTranslation(vehicle.node)
    local _, by, _ = getRotation(vehicle.baseNode)
    local _, vy, _ = getRotation(vehicle.node)

    local rotCorrection = by - vy
    return self.mission:loadVehicle(vehicle.filename, x, y, z, 0, rotCorrection + vehicle.rotation, true, 0, Vehicle.PROPERTY_STATE_NONE, vehicle.farmId, nil, nil, loadFinished, self, { queue })
end
