----------------------------------------------------------------------------------------------------
-- StartState
----------------------------------------------------------------------------------------------------
-- Purpose: Start state for the state machine.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class StartState
---@field public id number
StartState = {}

StartState.COUNT_DOWN_TIME = 10000 --ms
StartState.COUNT_DOWN_THRESHOLD = 5000 --ms

local StartState_mt = Class(StartState, AbstractState)

---Creates a new start state.
---@param id number
---@param stateMachine FSM
---@param trafficLight TrafficLight
---@param custom_mt table
---@return StartState
function StartState:new(id, stateMachine, trafficLight, hud, custom_mt)
    local self = AbstractState:new(id, stateMachine, custom_mt or StartState_mt)

    self.trafficLight = trafficLight
    self.hud = hud

    self.timer = Timer:new(StartState.COUNT_DOWN_TIME, function()
        self:startSuperCup()
    end)

    return self
end

---Called on delete.
function StartState:delete()
    StartState:superClass().delete(self)
end

---Called on enter.
function StartState:onEnter(runVehicle)
    StartState:superClass().onEnter(self)

    self.hud:setIsActive(false)
    self.runVehicle = runVehicle

    self.trafficLight:reset()
    self.timer:start()
end

function StartState:onExit()
    StartState:superClass().onExit(self)

    if self.timer:isRunning() then
        self.timer:clear()
    end
end

function StartState:onUpdate(dt)
    if self.timer:isRunning() then
        local elapsedTime = self.timer:getElapsedTime()

        if elapsedTime < StartState.COUNT_DOWN_THRESHOLD then
            local seconds = math.floor((elapsedTime * 0.001) + 1)
            if seconds ~= self.lastText then
                self.lastText = seconds

                if seconds == 3 then
                    self.trafficLight:next() -- to ORANGE.
                end
            end
        else
            self.lastText = "Get ready..."
        end

        self.hud:drawTextElement({ x = 0.5, y = 0.85, textSize = 0.04, text = tostring(self.lastText), bold = true })
    end
end

---Transition to next state.
function StartState:startSuperCup()
    self.trafficLight:next() -- to GREEN.
    self.stateMachine:transition(SuperCup.STATES.RUN, { self.runVehicle })
end
